/*
 * Copyright (C) Fondazione Istituto Italiano di Tecnologia
 *
 * Licensed under either the GNU Lesser General Public License v3.0 :
 * https://www.gnu.org/licenses/lgpl-3.0.html
 * or the GNU Lesser General Public License v2.1 :
 * https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 * at your option.
 */

#ifndef IDYNTREE_LINEAR_MOTION_VECTOR_3_H
#define IDYNTREE_LINEAR_MOTION_VECTOR_3_H

#warning <iDynTree/Core/LinearMotionVector3.h> is deprecated. Please use GeomVector3 and <iDynTree/Model/GeomVector3.h>.

#include <iDynTree/Core/GeomVector3.h>

#endif /* IDYNTREE_LINEAR_MOTION_VECTOR_3_H */
