function s=getLocation(obj)
% Return location of MoxUnitTestCase instance
%
% s=str(obj)
%
% Input:
%   obj             MOxUnitTestNode instance.
%
% Output:
%   s               location of obj.
%
    s=obj.location;